import networkx as nx

# Function to get nodes that should not be visited again
def get_dontgo(G, visited):
    dontgo = []
    if len(visited) > 1:  # Dont care if only source in the visited list
        for index, node in enumerate(visited[:-1]):
            dontgo = dontgo + list(
                nx.neighbors(G, node)
            )  # Get all neighbors of previous node
    return dontgo


# USe this. removes unnecessary paths also
def all_simple_paths(G, source, target):
    visited = [source]
    stack = [iter(G[source])]
    while stack:
        children = stack[-1]
        child = next(children, None)
        dontgo = get_dontgo(G, visited)
        if child is None:
            stack.pop()
            visited.pop()
        elif child not in dontgo:
            if child == target:
                yield visited + [target]
            elif child not in visited:
                visited.append(child)
                stack.append(iter(G[child]))
                
# Function to get all minimal paths between source and destination
def minimalpaths(G, src, dst):
    return list(all_simple_paths(G, src, dst))


from .sets import *
from .availability import *


__all__ = [
    "minimalcuts",
    "minimalcuts_optimized",
    "minimalpaths",
    'to_boolean_expression',
    'eval_single_pair',
    'eval_topology',
    'eval_avail_pyrbd',
    'eval_avail_pyrbd_multithreading',
    'eval_avail_pyrbd_multiprocessing'
]


from .engine import to_boolean_expression, eval_single_pair, eval_topology

def mcs_to_boolexpr(G, src, dst):
    """Convert a Minimal Cut Set (MCS) to a Boolean expression."""
    return to_boolean_expression(G, src, dst, 'mcs')

def eval_avail_mcs(G, A_dict, src, dst):
    """Evaluate the availability of a Minimal Cut Set (MCS) for a single source-destination pair."""
    return eval_single_pair(G, A_dict, src, dst, 'mcs')

def eval_avail_topo_mcs(G, A_dict):
    """Evaluate the availability of a Minimal Cut Set (MCS) for the entire topology."""
    return eval_topology(G, A_dict, 'mcs')

def eval_avail_topo_mcs_parallel(G, A_dict):
    """Evaluate the availability of a Minimal Cut Set (MCS) for the entire topology in parallel."""
    return eval_topology(G, A_dict, 'mcs', parallel=True)


from .evaluator import evaluate_availability
from .datasets import *
from .algorithms import minimalcuts_optimized, minimalpaths, minimalcuts
from .utils import relabel_graph_A_dict




__all__ = [
    'evaluate_availability',
    'read_graph',
    'read_mincutset',
    'read_pathset',
    'save_mincutset',
    'save_pathset',
    'save_boolean_expression_from_mincutset',
    'save_boolean_expression_from_pathset',
    'save_boolean_expression_from_sdp',
    'save_availability',
    'minimalcuts',
    'minimalcuts_optimized',
    'minimalpaths',
    'relabel_graph_A_dict',
    'to_link_graph',
    ]

__version__ = "0.1.0"


import networkx as nx

# Relabel the nodes of G and A_dic
def relabel_graph_A_dict(G, A_dic):
    # Get the nodes of G
    nodes = list(G.nodes())
    # Create a mapping of the nodes to new labels
    relabel_mapping = {nodes[i]: i + 1 for i in range(len(nodes))}
    # Relabel the nodes of G
    G_relabel = nx.relabel_nodes(G, relabel_mapping)
    # Relabel the nodes in A_dict
    A_dic = {relabel_mapping[node]: value for node, value in A_dic.items()}
    return G_relabel, A_dic, relabel_mapping

# relabel boolean expression back and convert the boolean expression to a mathematical expression
def relabel_boolexpr_to_str(bool_expr):
    expr = "["
    bool_expr_len = len(bool_expr)
    for num_list in bool_expr:
        expr += "["
        list_len = len(num_list)
        for num in num_list:
            if num == -1:
                expr += "0"
            elif num > 0:
                expr += str(num - 1)
            elif num < 0:
                expr += str(num + 1)
            list_len -= 1
            if list_len > 0:
                expr += " * "
        expr += "]"
        bool_expr_len -= 1
        if bool_expr_len > 0:
            expr += " + "
    expr += "]"
    return expr

def sdp_boolexpr_to_str(sdp_lst_lst):
    sdp_str = ""
    sdp_lst_lst_len = len(sdp_lst_lst)
    for sdp_lst in sdp_lst_lst:
        sdp_lst_len = len(sdp_lst)
        sdp_str += "["
        for sdp in sdp_lst:

            if sdp.isComplementary():
                sdp_str += "-["
            else:
                sdp_str += "["

            sdp_elem_len = len(sdp.getSet())
            for sdp_elem in sdp.getSet():
                if sdp_elem_len == 1:
                    sdp_str += f"{sdp_elem}"
                else:
                    sdp_str += f"{sdp_elem} * "
                sdp_elem_len -= 1
            sdp_str += "]"

            sdp_lst_len -= 1

            if sdp_lst_len > 0:
                sdp_str += " * "

        sdp_str += "]"
        sdp_lst_lst_len -= 1
        if sdp_lst_lst_len > 0:
            sdp_str += " + "

    return sdp_str.strip()


def sdp_boolexpr_length(sdp_lst_lst):
    expr_len = 0
  
    for sdp_lst in sdp_lst_lst:
        
        sdp_lst_len = len(sdp_lst)
        
        expr_len += sdp_lst_len

    return expr_len



"""
Consist of all availability evaluation methods from PyRBD
"""

import copy
import networkx as nx
from itertools import combinations
import multiprocessing
from ..sets import minimalcuts
from ...utils import relabel_graph_A_dict


# Find the most repeated node in cut sets with lowest cardinality
# Starting Node
def most_repeated_node(minimal_sets):
    min_cardinality = min(len(min_set) for min_set in minimal_sets)
    min_cardinality_sets = [
        min_set for min_set in minimal_sets if len(min_set) == min_cardinality
    ]

    nodes_count = {}
    for min_set in min_cardinality_sets:
        for node in min_set:
            if node in nodes_count:
                nodes_count[node] += 1
            else:
                nodes_count[node] = 1

    most_repeated_node = max(nodes_count, key=nodes_count.get)
    return most_repeated_node


# Check if there is direct path between s and t
def is_direct_path(G, src_, dst_):
    return nx.shortest_path_length(G, source=src_, target=dst_) == 1


# Check if there is path between source and target
def has_path(G, src_, dst_):
    return nx.has_path(G, source=src_, target=dst_)


# Modify the graph based on the existence or failure of the node
def manipulate_graph(graph, nodes):
    manipulated_graph = copy.deepcopy(graph)

    for node in nodes:
        node = int(node)
        # If node exist, remove node and add edge
        if node > 0:
            neighbors = list(manipulated_graph.neighbors(node))
            manipulated_graph.remove_node(node)
            for neighbor1 in neighbors:
                for neighbor2 in neighbors:
                    if neighbor1 != neighbor2 and not manipulated_graph.has_edge(
                        neighbor1, neighbor2
                    ):
                        manipulated_graph.add_edge(neighbor1, neighbor2)

        # If node failed, remove it
        elif node < 0:
            manipulated_graph.remove_node(node * (-1))
    return manipulated_graph


# Return a result dictionary mapping nodes to values: -1 for branching, 1 for direct paths, and 0 for failure.
def evaluate_nodes(graph, source, target, nodes_to_evaluate):
    result = {}  # Initializes an empty dictionary

    # Evaluate multiple node
    def evaluate_multiple_node(updated_graph, combined_nodes):

        try:
            result[f"{combined_nodes}"] = (
                1
                if is_direct_path(updated_graph, source, target)
                else -1 if has_path(updated_graph, source, target) else 0
            )
        except nx.NetworkXNoPath:
            result[f"{combined_nodes}"] = 0

    if isinstance(
        nodes_to_evaluate, int
    ):  # Single node case, only for the most repeated node

        with_node = manipulate_graph(graph, [nodes_to_evaluate])
        result[f"{nodes_to_evaluate}"] = (
            1
            if is_direct_path(with_node, source, target)
            else -1 if has_path(with_node, source, target) else 0
        )
        without_node = manipulate_graph(graph, [nodes_to_evaluate * (-1)])
        result[f"{nodes_to_evaluate * (-1)}"] = (
            1
            if is_direct_path(without_node, source, target)
            else -1 if has_path(without_node, source, target) else 0
        )

    elif isinstance(nodes_to_evaluate, list):  # Multiple nodes case

        for combination in nodes_to_evaluate:
            updated_graph = manipulate_graph(graph, combination)
            evaluate_multiple_node(updated_graph, combination)

    return result


# Returns the total availability of the system
def availability_evaluation(
    one_results, zero_results, component_data, source_availability, target_availability
):
    if len(one_results) > len(zero_results):
        if len(zero_results) != 0:

            unavailability = evaluate_expression(zero_results, component_data)
            availability = 1 - unavailability

            return availability * source_availability * target_availability
        else:
            availability = evaluate_expression(one_results, component_data)

            return availability * source_availability * target_availability

    elif len(one_results) < len(zero_results):
        if len(one_results) != 0:

            availability = evaluate_expression(one_results, component_data)

            return availability * source_availability * target_availability
        else:
            unavailability = evaluate_expression(zero_results, component_data)
            availability = 1 - unavailability

            return availability * source_availability * target_availability

    else:

        availability = evaluate_expression(one_results, component_data)

        return availability * source_availability * target_availability


# Substitute each node with its availability value
def value_substitution(char, component_data):
    char = int(char)
    if char < 0:
        return 1 - component_data.get(char * (-1))
    elif char > 0:
        return component_data.get(char)


# Calculates the total value of a list of words based on substitution values.
def evaluate_expression(results, component_data):
    total_value = 0
    for word in results:
        item_list = eval(word)
        word_value = 1
        for char in item_list:
            substituted_value = value_substitution(char, component_data)
            word_value *= substituted_value
        total_value += word_value
    return total_value


# A methods to evaluate the topology
def process_topology(G, source_node, target_node, A_dic, mincutsets=None):
    data = []
    availabilities = {}  # Avalibilities

    # Check if there is no direct connection between the source and target nodes with one hop
    if nx.shortest_path_length(G, source=source_node, target=target_node) > 1:

        if mincutsets is None:
            # Find all minimal cut sets
            all_cut_sets = minimalcuts(G, source_node, target_node)
        else:
            all_cut_sets = mincutsets

        # Find the most repeated node in the minimal cut sets
        # Starting node
        most_repeated = most_repeated_node(all_cut_sets)

        # Initializing two lists to store nodes with values 1: direct path and 0: no path
        zero_results = []
        one_results = []

        # Evaluate first node
        evaluate_result = evaluate_nodes(
            G, source_node, target_node, int(most_repeated)
        )

        # List to store unique nodes from cut sets excluding the most repeated node
        unique_nodes = []

        # Iterating through each cut set
        for cut_set in all_cut_sets:
            for node in cut_set:
                if node != most_repeated and node not in unique_nodes:
                    unique_nodes.append(node)

        previous_evaluation_nodes = set()
        # Iterating over unique nodes
        for node in unique_nodes:
            updated_evaluation_result = []

            # First iteration
            if not previous_evaluation_nodes:
                for key, value in evaluate_result.items():
                    if value == -1:
                        updated_evaluation_result.append([key, str(node)])
                        updated_evaluation_result.append([key, str(int(node * (-1)))])

                    iterated_nodes = evaluate_nodes(
                        G, source_node, target_node, updated_evaluation_result
                    )

                # Extract the keys from the evaluation result ( most repeated node evaluation )
                nodes = list(evaluate_result.keys())

                # Loop through the nodes and check if they are in the one or zero result
                for node in nodes:
                    if evaluate_result[node] == 1:
                        one_results.append(f"['{node}']")
                    elif evaluate_result[node] == 0:
                        zero_results.append(f"['{node}']")

                for key, value in iterated_nodes.items():
                    if value == -1:
                        previous_evaluation_nodes.add(key)
                    if value == 1:
                        one_results.append(key)
                    if value == 0:
                        zero_results.append(key)

            else:
                new_evaluation_nodes = set()
                # Using the evaluate_nodes from the previous iteration

                for s in previous_evaluation_nodes:
                    new_set = eval(
                        s
                    )  # Converting string representation of set to actual set
                    new_set2 = eval(s)
                    new_set.append(str(node))
                    new_set2.append(str(int(node * (-1))))
                    updated_evaluation_result.append(new_set)
                    updated_evaluation_result.append(new_set2)
                    iterated_nodes = evaluate_nodes(
                        G, source_node, target_node, updated_evaluation_result
                    )

                for key, value in iterated_nodes.items():
                    if value == -1:
                        new_evaluation_nodes.add(key)
                    if value == 1:
                        one_results.append(key)
                    if value == 0:
                        zero_results.append(key)
                previous_evaluation_nodes = new_evaluation_nodes

        combined_results = one_results + zero_results

        if isinstance(A_dic, dict):

            source_availability = A_dic.get(source_node)
            target_availability = A_dic.get(target_node)

            # Construct component_data with node availabilities
            component_data = {n: A_dic.get(n, A_dic.values()) for n in G.nodes()}
            # Calculate availability for each component value
            availabilities = availability_evaluation(
                one_results,
                zero_results,
                component_data,
                source_availability,
                target_availability,
            )
            # print('Overall Availability :', availabilities)
            data.append((source_node, target_node, availabilities))

        elif isinstance(A_dic, list):

            for component_value in A_dic:
                component_data = {node: component_value for node in G.nodes()}
                # source and target availabilities for the current component value
                source_availability = component_value
                target_availability = component_value
                availabilities[component_value] = availability_evaluation(
                    one_results,
                    zero_results,
                    component_data,
                    source_availability,
                    target_availability,
                )
                # print('Avalibility with flow avalibilities', component_value, 'is', availabilities[component_value])

            data.append(
                (source_node, target_node, *[availabilities[val] for val in A_dic])
            )
        else:
            print("Invalid input format")
        # data.append((source_node, target_node, *[availabilities[val] for val in A_dic]))
        return data, combined_results

    else:
        if isinstance(A_dic, dict):
            source_availability = A_dic.get(source_node)
            target_availability = A_dic.get(target_node)
            availabilities = source_availability * target_availability
            # print('Overall Availability :', availabilities)
            data.append((source_node, target_node, availabilities))

        elif isinstance(A_dic, list):
            for component_value in A_dic:
                source_availability = component_value
                target_availability = component_value
                availabilities[component_value] = (
                    source_availability * target_availability
                )
                # print('Avalibility with flow avalibilities', component_value, 'is', availabilities[component_value])

            data.append(
                (source_node, target_node, *[availabilities[val] for val in A_dic])
            )

        return data, []


# The main function to calculate the availability of the system
def eval_avail_pyrbd(G, src, dst, A_dict):
    """Evaluate the availability of a system using the PyRBD algorithm.

    Parameters:
    G (networkx.Graph): The graph representing the system.
    src (int): The source node.
    dst (int): The target node.
    A_dict (dict or list): A dictionary or list containing the availability values for each node.

    Returns:
    tuple: A tuple containing the result and combined results.
    """
    # Relabel the graph and availability dictionary
    G_relabel, A_dict_relabeled, relabel_mapping = relabel_graph_A_dict(G, A_dict)
    
    # Ensure src and dst are in the relabeled graph
    src_relabel = relabel_mapping[src]
    dst_relabel = relabel_mapping[dst]

    # Relabel the nodes of G
    G = nx.relabel_nodes(G, relabel_mapping)

    result, _ = process_topology(G_relabel, src_relabel, dst_relabel, A_dict_relabeled)

    return (src, dst, result[0][2])

def eval_avail_topo_pyrbd(G, A_dict):
    """Evaluate the availability of the entire topology using the PyRBD algorithm.

    Parameters:
    G (networkx.Graph): The graph representing the system.
    A_dict (dict or list): A dictionary or list containing the availability values for each node.

    Returns:
    List[tuple]: A list of tuples, each containing (source_node, target_node, availability).
    """
    # Relabel the graph and availability dictionary
    G_relabel, A_dict_relabeled, relabel_mapping = relabel_graph_A_dict(G, A_dict)
    
    # Get relabeled node pairs
    node_pairs = list(combinations(G_relabel.nodes(), 2))

    
    # Initialize result list
    results = []
    
    for src_relabel, dst_relabel in node_pairs:
        result, _ = process_topology(
            G_relabel, src_relabel, dst_relabel, A_dict_relabeled
        )
        results.append((src_relabel - 1, dst_relabel - 1, result[0][2]))

    return results

def eval_avail_topo_pyrbd_parallel(G, A_dict):
    """Evaluate the availability of the entire topology using the PyRBD algorithm in parallel.

    Parameters:
    G (networkx.Graph): The graph representing the system.
    A_dict (dict or list): A dictionary or list containing the availability values for each node.

    Returns:
    List[tuple]: A list of tuples, each containing (source_node, target_node, availability).
    """
    
    # Relabel the graph and availability dictionary
    G_relabel, A_dict_relabeled, relabel_mapping = relabel_graph_A_dict(G, A_dict)
    
    # Get relabeled node pairs
    node_pairs = list(combinations(G_relabel.nodes(), 2))
    
    # Get the number of CPUs available
    num_cpus = multiprocessing.cpu_count()
    
    # Create a multiprocessing pool
    pool = multiprocessing.Pool(processes=num_cpus)

    results = pool.starmap(
        process_topology, [(G_relabel, pair[0], pair[1], A_dict_relabeled) for pair in node_pairs]
    )

    pool.close()
    pool.join()

    result_list, combined_results = zip(*results)

    results = []
    for i in range(len(node_pairs)):
        results.append((node_pairs[i][0] - 1, node_pairs[i][1] - 1, result_list[i][0][2]))

    return results



from .engine import to_boolean_expression, eval_single_pair, eval_topology

def pathset_to_boolexpr(G, src, dst):
    """Convert a pathset to a Boolean expression."""
    return to_boolean_expression(G, src, dst, 'pathset')

def eval_avail_pathset(G, A_dict, src, dst):
    """Evaluate the availability of a pathset for a single source-destination pair."""
    return eval_single_pair(G, A_dict, src, dst, 'pathset')

def eval_avail_topo_pathset(G, A_dict):
    """Evaluate the availability of a pathset for the entire topology."""
    return eval_topology(G, A_dict, 'pathset')

def eval_avail_topo_pathset_parallel(G, A_dict):
    """Evaluate the availability of a pathset for the entire topology in parallel."""
    return eval_topology(G, A_dict, 'pathset', parallel=True)



from engine import to_boolean_expression, eval_single_pair, eval_topology

def sdp_to_boolexpr(G, src, dst):
    """Convert a Sum of Disjoint Products (SDP) to a Boolean expression."""
    return to_boolean_expression(G, src, dst, 'sdp')

def eval_avail_sdp(G, A_dict, src, dst):
    """Evaluate the availability of a Sum of Disjoint Products (SDP) for a single source-destination pair."""
    return eval_single_pair(G, A_dict, src, dst, 'sdp')

def eval_avail_topo_sdp(G, A_dict):
    """Evaluate the availability of a Sum of Disjoint Products (SDP) for the entire topology."""
    return eval_topology(G, A_dict, 'sdp')

def eval_avail_sdp_parallel(G, A_dict, src, dst):
    """Evaluate the availability of a Sum of Disjoint Products (SDP) for a single source-destination pair in parallel."""
    return eval_single_pair(G, A_dict, src, dst, 'sdp', parallel=True)

def eval_avail_topo_sdp_parallel(G, A_dict):
    """Evaluate the availability of a Sum of Disjoint Products (SDP) for the entire topology in parallel."""
    return eval_topology(G, A_dict, 'sdp', parallel=True)



"""
PyRBD++ Minimal Demo
====================
This demo shows the bare‑minimum usage of the library:

1) Load a graph
2) Build a node-probability dict
3) Evaluate availability
   - single (src, dst)
   - full topology (all pairs)
   - sequential vs parallel

No custom layout, no colors, no extra utilities.
"""

import time
from pyrbd3 import read_graph, evaluate_availability, to_link_graph


def print_section_start(title: str):
    print("\n" + "=" * 70)
    print(title)

def print_section_end():
    print("=" * 70)

def main():
    # Load a sample graph
    print_section_start("Load graph")
    topo = "Germany_17"
    G, _, _ = read_graph(f"topologies/{topo}", topo)
    print(f"Loaded {topo}: |V|={len(G.nodes())}, |E|={len(G.edges())}")

    # Build node probability dict (uniform p=0.9 as example)
    node_prob = {node: 0.9 for node in G.nodes()}
    first_five = list(node_prob.items())[:5]
    print("Sample Node Probabilities (first 5):", {n: p for n, p in first_five})
    print_section_end()

    # Single pair evaluation with several algorithms
    src, dst = 0, 1
    print_section_start(f"Single pair availability: src={src}, dst={dst}")
    # PyRBD
    pyrbd_res = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="pyrbd")
    print("PyRBD Result:", pyrbd_res)
    # PyRBD++ MCS
    pyrbdpp_mcs_res = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="mcs")
    print("PyRBD++ MCS Result:", pyrbdpp_mcs_res)
    # PyRBD++ Pathset
    pyrbdpp_pathset_res = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="pathset")
    print("PyRBD++ Pathset Result:", pyrbdpp_pathset_res)
    # PyRBD++ SDP
    pyrbdpp_sdp_res = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="sdp")
    print("PyRBD++ SDP Result:", pyrbdpp_sdp_res)
    print_section_end()
    
    # Full topology (all pairs), sequential
    print_section_start("All pairs (PyRBD++ SDP, sequential)")
    t0 = time.time()
    res_all = evaluate_availability(G, node_prob, algorithm="sdp")
    print("Sample Results (first 2 pairs):", res_all[:2])
    dt = time.time() - t0
    print(f"Time: {dt:.4f} s")
    print_section_end()

    # Full topology (all pairs), parallel
    print_section_start("All pairs (PyRBD++ SDP, parallel)")
    t0 = time.time()
    res_all_p = evaluate_availability(G, node_prob, algorithm="sdp", parallel=True)
    print("Sample Results (first 2 pairs):", res_all_p[:2])
    dt = time.time() - t0
    print(f"Time: {dt:.4f} s")
    print_section_end()
    
    # Link-counted evaluation example
    print_section_start("Link-counted evaluation example")
    edge_prob = {edge: 0.95 for edge in G.edges()}
    t0 = time.time()
    res_link_counted = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="sdp", count_link=True, edge_prob=edge_prob)
    print("Link-counted Result:", res_link_counted)
    dt = time.time() - t0
    print(f"Time: {dt:.4f} s")
    print_section_end()
    
    # Link-counted full topology, parallel
    print_section_start("All pairs link-counted (PyRBD++ SDP, parallel)")
    t0 = time.time()
    res_all_link = evaluate_availability(G, node_prob, algorithm="sdp", parallel=True, count_link=True, edge_prob=edge_prob)
    print("Sample Results (first 2 pairs):", res_all_link[:2])
    dt = time.time() - t0
    print(f"Time: {dt:.4f} s")
    print_section_end()
    
    print("Customize algorithms, src/dst, probabilities, or topology as needed.")


if __name__ == "__main__":
    main()


import os
import pickle as pkl
import pandas as pd
import ast
from itertools import combinations
from tqdm import tqdm
import numpy as np

from .algorithms.sets import minimalcuts_optimized, minimalpaths
from .utils import relabel_boolexpr_to_str, sdp_boolexpr_to_str, sdp_boolexpr_length


import pyrbd3._core.pyrbd3_cpp as cpp


# Load data from a pickle file
def read_graph(directory, top, file_path=None):
    if file_path:
        with open(file_path, "rb") as handle:
            f = pkl.load(handle)
    else:
        with open(os.path.join(directory, "Pickle_" + top + ".pickle"), "rb") as handle:
            f = pkl.load(handle)
    G = f[0]
    pos = f[1]
    label = f[2]

    return G, pos, label

# Read mincutset from a csv file
def read_mincutset(directory, top):
    """Read mincutset and lengths from a csv file.

    Args:
        directory (str): The directory containing the CSV file.
        top (str): The name of the topology.

    Returns:
        tuple: A tuple containing two lists - the mincutsets and their lengths.
    """
    # Read the mincutset from a csv file
    df = pd.read_csv(os.path.join(directory, "Mincutset_" + top + ".csv"))

    # Convert the 'mincutsets' column from string representation to actual list
    df["mincutsets"] = df["mincutsets"].apply(ast.literal_eval)

    # Extract the 'mincutsets' column as a list
    mincutsets = df["mincutsets"].values.tolist()

    # Extract the 'length' column as a list
    lengths = df["length"].values.tolist()

    return mincutsets, lengths


# Read pathset from a csv file
def read_pathset(directory, top):
    """Read pathset and lengths from a csv file.

    Args:
        directory (str): The directory containing the CSV file.
        top (str): The name of the topology.

    Returns:
        tuple: A tuple containing two lists - the pathsets and their lengths.
    """
    # Read the pathset from a csv file
    df = pd.read_csv(os.path.join(directory, "Pathset_" + top + ".csv"))

    # Convert the 'pathsets' column from string representation to actual list
    df["pathsets"] = df["pathsets"].apply(ast.literal_eval)

    # Extract the 'pathsets' column as a list
    pathsets = df["pathsets"].values.tolist()

    # Extract the 'length' column as a list
    lengths = df["length"].values.tolist()

    return pathsets, lengths


# Evaluate the mincutset and save it to a csv file
def save_mincutset(directory, top):
    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Get all node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Initialize a list to store mincutsets data
    mincutsets_data = []

    # Iterate through each pair of nodes and find the minimal cut sets
    for src, dst in tqdm(node_pairs, desc=f"Saving Mincutsets for {top}", leave=False):
        mincutset = minimalcuts_optimized(G, src, dst)
        if mincutset:
            mincutsets_data.append(
                {
                    "source": src,
                    "target": dst,
                    "mincutsets": repr(mincutset),
                    "length": len(mincutset),
                }
            )

    # Save mincutsets to CSV
    mincutsets_df = pd.DataFrame(mincutsets_data)
    mincutsets_df.to_csv(os.path.join(directory, f"Mincutset_{top}.csv"), index=False)


# Evaluate the pathset and save it to a csv file
def save_pathset(directory, top):
    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Get all node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Initialize a list to store pathsets data
    pathsets_data = []

    # Iterate through each pair of nodes and find the simple paths
    for src, dst in tqdm(node_pairs, desc=f"Saving Pathsets for {top}", leave=False):
        pathset = minimalpaths(G, src, dst)
        if pathset:
            # Sort paths by length and for the same length, sort by lexicographical order
            pathset = sorted(pathset, key=lambda x: (len(x), x))
            pathsets_data.append(
                {
                    "source": src,
                    "target": dst,
                    "pathsets": repr(pathset),
                    "length": len(pathset),
                }
            )

    # Save pathsets to CSV
    pathsets_df = pd.DataFrame(pathsets_data)
    pathsets_df.to_csv(os.path.join(directory, f"Pathset_{top}.csv"), index=False)


# Evaluate the boolean expression from the mincutsets and save it to a csv file
def save_boolean_expression_from_mincutset(directory, top):
    # Read the mincutset from the csv file
    mincutsets, _ = read_mincutset(directory, top)

    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Create node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Initialize a list to store boolean expressions data
    boolean_expressions_data = []

    # Iterate through each mincutset and evaluate the boolean expression
    for i, (src, dst) in tqdm(
        enumerate(node_pairs),
        desc=f"Saving Boolean Expressions from MCS for {top}",
        leave=False,
        total=len(node_pairs),
    ):
        # Relabel mincutset by adding 1 to ensure no 0 in the set
        mincutsets[i] = [[node + 1 for node in cutset] for cutset in mincutsets[i]]

        # Build the boolean expression from the mincutset
        expression = cpp.mcs.to_probaset(src, dst, mincutsets[i])

        # Append the boolean expression data
        boolean_expressions_data.append(
            {
                "source": src,
                "target": dst,
                "boolean_expression": relabel_boolexpr_to_str(expression),
                "length": len(expression),
            }
        )

    # Save boolean expressions to CSV
    boolean_expressions_df = pd.DataFrame(boolean_expressions_data)
    boolean_expressions_df.to_csv(
        os.path.join(directory, f"BoolExprMCS_{top}.csv"), index=False
    )


# Evaluate the boolean expression from the pathset and save it to a csv file
def save_boolean_expression_from_pathset(directory, top):
    # Read the pathset from the csv file
    pathsets, _ = read_pathset(directory, top)

    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Create node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Initialize a list to store boolean expressions data
    boolean_expressions_data = []

    # Iterate through each pathset and evaluate the boolean expression
    for i, (src, dst) in tqdm(
        enumerate(node_pairs),
        desc=f"Saving Boolean Expressions from Pathset for {top}",
        leave=False,
        total=len(node_pairs),
    ):
        # Relabel pathset by adding 1 to ensure no 0 in the set
        pathsets[i] = [[node + 1 for node in path] for path in pathsets[i]]

        # Build the boolean expression from the pathset
        expression = cpp.pathset.to_probaset(src, dst, pathsets[i])

        # Append the boolean expression data
        boolean_expressions_data.append(
            {
                "source": src,
                "target": dst,
                "boolean_expression": relabel_boolexpr_to_str(expression),
                "length": len(expression),
            }
        )

    # Save boolean expressions to CSV
    boolean_expressions_df = pd.DataFrame(boolean_expressions_data)
    boolean_expressions_df.to_csv(
        os.path.join(directory, f"BoolExprPS_{top}.csv"), index=False
    )


# Evaluate the boolean expression from the SDP and save it to a csv file
def save_boolean_expression_from_sdp(directory, top):
    # Read the mincutset from the csv file
    pathsets, _ = read_pathset(directory, top)

    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Create node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Initialize a list to store boolean expressions data
    boolean_expressions_data = []

    for i, (src, dst) in tqdm(
        enumerate(node_pairs),
        desc=f"Saving Boolean Expressions from SDP for {top}",
        leave=False,
        total=len(node_pairs),
    ):
        # Relabel pathset by adding 1 to ensure no 0 in the set
        pathsets[i] = [[node + 1 for node in path] for path in pathsets[i]]

        # Build the SDP set from the pathset
        expression = cpp.sdp.to_sdp_set_parallel(src, dst, pathsets[i])

        # Convert the SDP set to a string representation and calculate the expression length
        bool_expr_str = sdp_boolexpr_to_str(expression)

        # Append the result to the list
        boolean_expressions_data.append(
            {
                "src": src,
                "dst": dst,
                "boolean_expression": bool_expr_str,
                "length": sdp_boolexpr_length(expression),
            }
        )

    # Save the boolean expressions to a CSV file
    boolean_expressions_df = pd.DataFrame(boolean_expressions_data)
    boolean_expressions_df.to_csv(
        os.path.join(directory, f"BoolExprSDP_{top}.csv"), index=False
    )


# Evaluate the availability of a network based on the given parameters
def save_availability(directory, top):
    from .evaluator import evaluate_availability

    """
    Save the availability of the topology to a CSV file.
    The result is with all nodes availablity set to 0.9, 0.99, 0.999 and 0.9999.
    Args:
        directory (str): The directory where the topology is located.
        top (str): The name of the topology.
    """
    # Read the graph from the pickle file
    G, _, _ = read_graph(directory, top)

    # Create node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Create nodes probabilities dictionary
    node_proba = {
        "0.9": [0.9] * len(G.nodes()),
        "0.99": [0.99] * len(G.nodes()),
        "0.999": [0.999] * len(G.nodes()),
        "0.9999": [0.9999] * len(G.nodes()),
    }

    # Initialize a list to store availability data
    result_data = {
        "source": [pair[0] for pair in node_pairs],
        "target": [pair[1] for pair in node_pairs],
    }

    # Iterate through each probability and evaluate availability
    for prob, values in tqdm(
        node_proba.items(), desc=f"Saving Availability for {top}", leave=False
    ):
        nodes_probabilities = dict(zip(G.nodes(), values))

        # Evaluate availability for all pairs using SDP algorithm
        eval_result = evaluate_availability(
            G, nodes_probabilities, algorithm="sdp", parallel=True
        )
        avail = [t[2] for t in eval_result]

        # Save the results for this probability
        result_data[f"{prob} Availability"] = avail

    # Save availability data to CSV
    availability_df = pd.DataFrame(result_data)
    availability_df.to_csv(
        os.path.join(directory, f"Availability_{top}.csv"),
        index=False,
        float_format="%.9f",
    )


def evaluate_boolean_expressions(directory, topologies, node_prob):
    # Read the graph
    G, _, _ = read_graph(directory, topologies)

    # Check if all nodes has probabilities assigned
    if not all(node in node_prob for node in G.nodes()):
        raise ValueError("Not all nodes have probabilities assigned.")

    # Read the boolean expressions
    df = pd.read_csv(os.path.join(directory, f"BoolExprSDP_{topologies}.csv"))

    # Create node pairs
    node_pairs = list(combinations(G.nodes(), 2))

    # Evaluate availability for each node pair
    for i, (src, dst) in tqdm(enumerate(node_pairs), desc=f"Evaluating Boolean Expressions for {topologies}", leave=False):
        # Get the boolean expression for the current node pair
        bool_exprs = df.iloc[i]["boolean_expression"]
        print(bool_exprs)
        
        
def to_link_graph(G, A_dict, edge_prob):
    """Convert a NetworkX graph to a link counted graph representation.
    
    If 0 - 1 is connected in G, then in Link_G, there will be a node (0,1), 0 - (0,1) - 1.
    We keep the original node numbers and add new nodes for each edge.
    The edge_mapping dictionary maps each original edge to its corresponding new node in Link_G.

    Args:
        G (nx.Graph): The input graph.
        A_dict (dict): A dictionary mapping nodes to their availability.
        edge_prob (dict): A dictionary mapping edges to their availability.

    Returns:
        Link_G (nx.Graph): The link counted graph representation.
        A_dict_new (dict): Updated availability dictionary including new link nodes.
    """        
    # Convert the link counted graph
    link_G = G.copy()
    A_dict_new = A_dict.copy()
    
    for i, (u, v) in enumerate(G.edges()):
        new_node = len(G.nodes()) + i

        # add new node and connect it to u and v
        link_G.add_node(new_node)
        link_G.add_edge(u, new_node)
        link_G.add_edge(new_node, v)

        # remove old connection
        link_G.remove_edge(u, v)
        
        # update A_dict
        A_dict_new[new_node] = edge_prob.get((u, v), edge_prob.get((v, u), 0.0))

    return link_G, A_dict_new


def dataset_preparation(directory, topologies):
    """
    Prepare the dataset by saving pathsets, mincutsets, and boolean expressions.
    Args:
        directory (str): The directory where the datasets will be saved.
        topologies (list): A list of topology names to process.
    """
    for top in topologies:
        directory = f"topologies/{top}"
        save_pathset(directory, top)
        save_boolean_expression_from_pathset(directory, top)
        save_boolean_expression_from_sdp(directory, top)
        save_mincutset(directory, top)
        save_boolean_expression_from_mincutset(directory, top)



from itertools import combinations
from loguru import logger

from ...algorithms.sets import minimalcuts_optimized, minimalpaths
from ...utils import relabel_graph_A_dict, relabel_boolexpr_to_str, sdp_boolexpr_to_str
from .pyrbd import eval_avail_pyrbd, eval_avail_topo_pyrbd, eval_avail_topo_pyrbd_parallel
import pyrbd3._core.pyrbd3_cpp as cpp

# Algorithm Configuration
ALGORITHM_CONFIG = {
    'mcs': {
        'cpp_module': 'mcs',
        'problem_set_func': minimalcuts_optimized,
        'bool_expr_func': relabel_boolexpr_to_str,
        'to_set_func': 'to_probaset',
    },
    'pathset': {
        'cpp_module': 'pathset', 
        'problem_set_func': minimalpaths,
        'bool_expr_func': relabel_boolexpr_to_str,
        'to_set_func': 'to_probaset',
    },
    'sdp': {
        'cpp_module': 'sdp',
        'problem_set_func': minimalpaths,  # SDP is based on pathsets
        'bool_expr_func': sdp_boolexpr_to_str,
        'to_set_func': 'to_sdp_set',
    },
    'pyrbd': {
        'cpp_module': None,
        'problem_set_func': None,
        'bool_expr_func': None,
        'to_set_func': None,
    }
}

def to_boolean_expression(G, src, dst, algorithm='mcs'):
    """Convert a SUM of Disjoint Products (SDP) to a Boolean expression.
    
    Args:
        G (networkx.Graph): The graph representation of the topology.
        src (int): The source node index.
        dst (int): The destination node index.
        algorithm (str): The algorithm to use for generating the Boolean expression. Can be 'mcs', 'pathset', 'sdp', or 'pyrbd'.
    
    Returns:
        bool_expr (str): A string representation of the Boolean expression.
    """
    if algorithm not in ALGORITHM_CONFIG:
        raise ValueError(f"Unsupported algorithm: {algorithm}. Choose from {list(ALGORITHM_CONFIG.keys())}.")
    
    if algorithm == 'pyrbd':
        raise NotImplementedError("Boolean expression generation for PyRBD is not implemented.")
    
    # Get algorithm configuration
    config = ALGORITHM_CONFIG[algorithm]
    cpp_module = getattr(cpp, config['cpp_module'])
    
    # Relabel
    G_relabel, _, relabel_mapping = relabel_graph_A_dict(G, {})
    src_relabel = relabel_mapping[src]
    dst_relabel = relabel_mapping[dst]
    
    # Get problem sets
    problem_sets = config['problem_set_func'](G_relabel, src_relabel, dst_relabel)
    
    # Convert to probability set or SDP set
    to_set_func = getattr(cpp_module, config['to_set_func'])
    result_set = to_set_func(src_relabel, dst_relabel, problem_sets)
    
    # Convert to boolean expression
    return config['bool_expr_func'](result_set)

def eval_single_pair(G, A_dict, src, dst, algorithm='sdp', parallel=False, count_link=False, edge_prob=None):
    """Evaluate the availability of a source and destination pair in the topology using SDP.
    
    Args:
        G (networkx.Graph): The graph representation.
        A_dict (dict): A dictionary mapping nodes to their availability.
        src (int): The source node index.
        dst (int): The destination node index.
        algorithm (str): The algorithm to use for evaluation ('mcs', 'pathset', or 'sdp').
        parallel (bool): Whether to use parallel evaluation if available.
        count_link (bool): Whether to consider link availability.
        edge_prob (dict): A dictionary mapping edges to their availability if count_link is True.
    
    Raises:
        ValueError: If the specified algorithm does not support parallel evaluation.
    
    Returns:
        tuple: (src, dst, availability)
    """
    # Validate algorithm
    if algorithm not in ALGORITHM_CONFIG:
        raise ValueError(f"Unsupported algorithm: {algorithm}. Choose from {list(ALGORITHM_CONFIG.keys())}.")
    
    if algorithm == 'pyrbd':
        return eval_avail_pyrbd(G, src, dst, A_dict)
    
    if count_link and not edge_prob:
        raise ValueError("Edge probabilities must be provided when count_link is True.")
    
    # Get algorithm configuration
    config = ALGORITHM_CONFIG[algorithm]
    cpp_module = getattr(cpp, config['cpp_module'])
    
    # Handle link counting
    if count_link:
        # Convert to link graph
        from ...datasets import to_link_graph
        G, A_dict = to_link_graph(G, A_dict, edge_prob)

    # Relabel
    G_relabel, A_dict_relabeled, relabel_mapping = relabel_graph_A_dict(G, A_dict)
    src_relabel = relabel_mapping[src]
    dst_relabel = relabel_mapping[dst]
    
    # Get problem sets
    problem_sets = config['problem_set_func'](G_relabel, src_relabel, dst_relabel)
    
    # Choose evaluation method
    if parallel:
        if hasattr(cpp_module, 'eval_avail_parallel'):
            availability = cpp_module.eval_avail_parallel(
                src_relabel, dst_relabel, A_dict_relabeled, problem_sets
            )
        else:
            raise ValueError(f"Parallel evaluation not available for {algorithm} algorithm.")
    else:
        availability = cpp_module.eval_avail(
            src_relabel, dst_relabel, A_dict_relabeled, problem_sets
        )
    
    return (src, dst, availability)

def eval_topology(G, A_dict, algorithm='sdp', parallel=False, count_link=False, edge_prob=None):
    """Evaluate the availability for all pairs of nodes in the topology using SDP.

    Args:
        G (networkx.Graph): The graph representation.
        A_dict (dict): A dictionary mapping nodes to their availability.
        algorithm (str): The algorithm to use for evaluation ('mcs', 'pathset', or 'sdp').
        parallel (bool): Whether to use parallel evaluation if available.
        count_link (bool): Whether to consider link availability.
        edge_prob (dict): A dictionary mapping edges to their availability if count_link is True.
        
    Raises:
        ValueError: If the specified algorithm does not support parallel evaluation.
    
    Returns:
        List[tuple]: A list of tuples, each containing (src, dst, availability).
    """
    # Validate algorithm
    if algorithm not in ALGORITHM_CONFIG:
        raise ValueError(f"Unsupported algorithm: {algorithm}. Choose from {list(ALGORITHM_CONFIG.keys())}.")
    
    if algorithm == 'pyrbd':
        return eval_avail_topo_pyrbd(G, A_dict) if not parallel else eval_avail_topo_pyrbd_parallel(G, A_dict)
    
    if count_link and not edge_prob:
        raise ValueError("Edge probabilities must be provided when count_link is True.")
    
    # Get algorithm configuration
    config = ALGORITHM_CONFIG[algorithm]
    cpp_module = getattr(cpp, config['cpp_module'])
    
    # Handle link counting
    if count_link:
        # Convert to link graph
        from ...datasets import to_link_graph
        G, A_dict = to_link_graph(G, A_dict, edge_prob)
    
    # Relabel
    G_relabel, A_dict_relabeled, relabel_mapping = relabel_graph_A_dict(G, A_dict)
    reverse_mapping = {v: k for k, v in relabel_mapping.items()}
    
    # Get all pairs
    node_pairs = list(combinations(G_relabel.nodes(), 2))
    
    # Get all problem sets
    problem_sets_list = [
        config['problem_set_func'](G_relabel, src, dst)
        for src, dst in node_pairs
    ]

    # Choose evaluation method
    if parallel:
        if hasattr(cpp_module, 'eval_avail_topo_parallel'):
            availability_lst = cpp_module.eval_avail_topo_parallel(
                node_pairs, A_dict_relabeled, problem_sets_list
            )
        else:
            raise ValueError(f"Parallel evaluation not available for {algorithm} algorithm.")
    else:
        availability_lst = cpp_module.eval_avail_topo(
            node_pairs, A_dict_relabeled, problem_sets_list
        )
    
    # Relabel results
    return [
        (reverse_mapping[src], reverse_mapping[dst], availability)
        for src, dst, availability in availability_lst
    ]



import networkx as nx
from itertools import combinations, islice
import numpy as np
import math
from .pathsets import minimalpaths


def successpaths(H, source, target, weight="weight"):
    return list(nx.shortest_simple_paths(H, source, target, weight=weight))


# Function for Finding Minimal cut sets
def minimalcuts(H, src_, dst_, order=None):
    if order is None:
        order = math.ceil(len(H.nodes) / 2)
    
    # modified here
    paths = list(islice(nx.shortest_simple_paths(H, src_, dst_, weight="weight"), 2))
    minimal = []

    if [src_, dst_] in paths:
        minimal.append([src_])
        minimal.append([dst_])
    else:
        paths = successpaths(H, src_, dst_)
        pairs = np.array(H.nodes)
        pairs = pairs.tolist()

        pairs = [pair for pair in pairs if pair != src_ and pair != dst_]

        incidence = np.zeros([len(paths), len(pairs)])
        incidence_cols_name = pairs

        for x in range(len(paths)):
            for comp in pairs:
                if comp in paths[x]:
                    incidence[x, pairs.index(comp)] = 1

        firstpairs = []
        for k in range(1, order + 1):
            if incidence.shape[1] == 0:
                break

            all_ones = [i for i in range(incidence.shape[1]) if incidence[:, i].all()]
            if k == 1:
                firstpairs = [k for k in incidence_cols_name if k not in all_ones]
            for c in all_ones:
                minimal.append(incidence_cols_name[c])

            if k >= order:
                continue

            pairs = firstpairs
            newpairs = list(combinations(pairs, k + 1))

            newpairstodelete = []
            for i in newpairs:
                for j in minimal:
                    if isinstance(j, tuple):
                        if set(j).issubset(i):
                            newpairstodelete.append(i)
                            break
            newpairs = [i for i in newpairs if i not in newpairstodelete]

            incidence_ = np.zeros([len(paths), len(newpairs)])
            incidence_cols_name = newpairs
            for x in range(len(paths)):
                for y in range(len(newpairs)):
                    for comp in newpairs[y]:
                        if comp in paths[x]:
                            incidence_[x, y] = 1
                            break
            incidence = np.copy(incidence_)

    return minimal


# Optimized version of minimalcuts function
def minimalcuts_optimized(H, src_, dst_, order=None):
    # set the order of the minimal cut set, can be modified here
    if order is None:
        order = math.ceil(len(H.nodes) / 2)

    # save the minimal cut sets
    minimal = []

    if nx.shortest_path_length(H, source=src_, target=dst_) > 1:
        # get all paths from source to destination

        # this method is slower than the one below
        # paths = successpaths(H, src_, dst_)
        # paths = remove_unnecessary_loops(paths, H)
        paths = minimalpaths(H, src_, dst_)
    
        # get all nodes in the graph except source and destination
        nodes = np.array(H.nodes)
        nodes = nodes.tolist()
        valid_nodes = [n for n in nodes if n != src_ and n != dst_]

        # create origin incidence matrix: columns are all nodes except source and destination, rows are paths
        indices_origin = np.zeros([len(paths), len(valid_nodes)], dtype=bool)
        indices_columns_origin = valid_nodes

        # create a dictionary to map node to column index
        node_to_col = {node: i for i, node in enumerate(valid_nodes)}

        # create incidence matrix
        """
        for example: 
        src = 1, dst = 4, nodes = [1, 2, 3, 4]
        path1 = [1, 2, 3, 4], path2 = [1, 3, 4]

        indices_origin:
                 2     3  
        path1: true  true
        path2: false true
        """
        for x in range(len(paths)):
            for comp in valid_nodes:
                if comp in paths[x]:
                    indices_origin[x, valid_nodes.index(comp)] = True
        
        # set the current incidence matrix to origin incidence matrix, update the column names
        indices_current = np.copy(indices_origin)
        indices_columns_current = indices_columns_origin

        # find all true columns index
        all_true_index = np.all(indices_current, axis=0)
        
        # find all true columns name
        all_true = [
            indices_columns_current[j]
            for j, is_true in enumerate(all_true_index)
            if is_true
        ]

        # find the first pairs that are columns not all true
        firstpairs = [k for k in indices_columns_current if k not in all_true]

        # loop for finding minimal cut sets
        for k in range(1, order + 1):

            # break if there is no columns in the incidence matrix
            if indices_current.shape[1] == 0:
                break

            # find all true columns index
            all_true_index = np.all(indices_current, axis=0)
            all_true = [
                indices_columns_current[j]
                for j, is_true in enumerate(all_true_index)
                if is_true
            ]

            # add all true columns to minimal cut sets
            for elem in all_true:
                minimal.append((elem,) if not isinstance(elem, (list, tuple)) else elem)

            if k >= order:
                continue

            # find all combinations of k + 1 pairs
            all_combinations = set(combinations(firstpairs, k + 1))

            # put all first pairs into a set
            firstpairsets = set(firstpairs)

            # create a set to store the upper sets
            upperset = set()

            # finding all the upper sets of the current minimal cut sets
            for tup in minimal:
                # put the current minimal cut set into a set
                tupset = set(tup)
                
                # find the remainder of the first pairs and the current minimal cut set
                remainder = firstpairsets - tupset
                
                # find the number of nodes needed to add to the current minimal cut set
                need = k + 1 - len(tup)

                # if the number of nodes needed is 0, continue
                if need == 0:
                    continue

                # find all the combinations of the remainder nodes
                for c in combinations(remainder, need):

                    # add the current minimal cut set and the new combinations of remainder
                    # add the sorted tuple to the upper set
                    upperset.add(tuple(sorted(c + tup)))

            # find the new pairs by removing the upper sets of current minimal cut sets from all combinations
            newpairs = list(all_combinations - upperset)

            # create the new incidence matrix, columns are new pairs, rows are paths
            indices_current = np.zeros([len(paths), len(newpairs)], dtype=bool)
            indices_columns_current = newpairs
        
            # set the value of the new incidence matrix according to the original incidence matrix for each new pair
            for j, nodes in enumerate(newpairs):

                # find the columns index in the original incidence matrix of each node in new pair
                """
                for example:
                nodes = [2, 3]
                node_to_col = {2: 0, 3: 1}
                indices_tmp = [0, 1]
                """
                indices_tmp = [node_to_col[node] for node in nodes]

                # select all the columns in the original incidence matrix,
                # do the logical or operation for each row,
                # set the result to the new incidence matrix column
                """
                for example:
                indices_current:
                        [2, 3]  =    2   OR  3
                path1:   true   =  true  OR true
                path2:   true   =  false OR true
                """
                indices_current[:, j] = np.any(indices_origin[:, indices_tmp], axis=1)

    # convert the minimal cut sets to list of lists
    minimal = [list(i) for i in minimal]

    # sort the minimal cut sets by length and lexicographically
    minimal = sorted(minimal, key=lambda x: (len(x), x))

    # add source and destination to the minimal cut sets
    minimal.insert(0, [dst_])
    minimal.insert(0, [src_])

    return minimal



from .datasets import read_graph
from .algorithms.availability import eval_single_pair, eval_topology

def evaluate_availability(
    graph_or_filepath,
    nodes_probabilities,
    src=None,
    dst=None,
    algorithm="sdp",
    parallel=False,
    count_link=False,
    edge_prob=None,
):
    """Evaluate the availability of a network based on the given parameters.

    Args:
        graph_or_filepath (Union[str, nx.Graph]): Either a file path to the graph pickle data or a NetworkX graph object.
        nodes_probabilities (dict): A dictionary mapping node IDs to their availability probabilities.
        src (int, optional): The source node ID. Defaults to None.
        dst (int, optional): The destination node ID. Defaults to None.
        algorithm (str, optional): The algorithm to use for evaluation. Defaults to 'sdp'. Can be 'mcs', 'pathset', 'sdp', or 'pyrbd'.
        parallel (bool, optional): Whether to evaluate in parallel. Defaults to False.
        count_link (bool, optional): Whether to consider link availability. Defaults to False.
        edge_prob (dict, optional): A dictionary mapping edge tuples to their availability probabilities if count
    """
    # Read the graph from the specified directory and topology
    if isinstance(graph_or_filepath, str):
        G, _, _ = read_graph("", "", graph_or_filepath)
    elif hasattr(graph_or_filepath, "nodes") and hasattr(graph_or_filepath, "edges"):
        # Assume graph_or_filepath is a networkx.Graph object
        G = graph_or_filepath
    else:
        raise TypeError(
            "Invalid graph input. Provide a file path or a NetworkX graph object."
        )
    # Validate the nodes_probabilities
    if not isinstance(nodes_probabilities, dict):
        raise TypeError(
            "nodes_probabilities must be a dictionary mapping node IDs to probabilities."
        )
    if len(nodes_probabilities) != len(G.nodes()):
        raise ValueError(
            "nodes_probabilities must contain probabilities for all nodes in the graph."
        )

    # Validate source and destination nodes
    if src is not None and dst is not None:
        if src not in G.nodes() or dst not in G.nodes():
            raise ValueError(
                f"Source node {src} or destination node {dst} not found in the graph."
            )
        return eval_single_pair(G, nodes_probabilities, src, dst, algorithm, parallel, count_link, edge_prob)
    elif src is None and dst is None:
        return eval_topology(G, nodes_probabilities, algorithm, parallel, count_link, edge_prob)
    else:
        raise ValueError(
            "Both source and destination nodes must be specified or neither."
        )



from setuptools import setup, find_packages

setup(
    name="pyrbd3",
    version="1.0.0",
    packages=find_packages(include=["src/pyrbd3", "src/pyrbd3.*"]),
    install_requires=[
        "pandas",
        "networkx",
        "tqdm",
        "loguru"
    ],
)


from pyrbd3 import read_graph, evaluate_availability

topo = "Germany_17"
G, _, _ = read_graph(f"topologies/{topo}", topo)

node_prob = {node: 0.9 for node in G.nodes()}

src, dst = 0, 1

availability = evaluate_availability(G, node_prob, src=src, dst=dst, algorithm="sdp")

print(f"System availability from {src} to {dst} in {topo}: {availability:.6f}")